"Boot script for closures.  Assumes current directory contains an image and changes and possibly a source file.  Assumes parent directory contains a directory ``changesets'' containing the various change sets.  Click inside the first parenthesis to select the first doit and evaluate to load the bootstrap and recompile the system.  If you are prompted that a class variable in CompiledMethod is still used, answer that it is OK to move it to Undeclared.  If you're bootstrapping Croquet 1.0 expect one compilation error in CTerrainTool>>build.  It uses a future which isn't sent a message.  Once you have run the bootstrap read and obey the comment ``To eliminate the last of the old methods...'' at the end of this file.  Good luck!!"

(| sic previouslyUndeclared newlyUndeclared unboundMethods hasTweak isSqueak39 getLatest dot dotDot csdir loadCS canAndShouldDeleteGenerateWithUsing |
#(initFields voidMapCache) size. "Squash compiler warnings"

(sic := SmalltalkImage current)
	closeSourceFiles; "so copying the changes file will always work"
	saveChangesInFileNamed: (sic fullNameForChangesNamed: 'ClosureBootstrap.image');
	saveImageInFileNamed: (sic fullNameForImageNamed: 'ClosureBootstrap.image').

Undeclared isEmpty ifFalse:
	[Undeclared removeUnreferencedKeys].
previouslyUndeclared := Undeclared keys.

hasTweak := Smalltalk includesKey: #Tweak.
canAndShouldDeleteGenerateWithUsing := (MethodNode includesSelector: #generateWith:using:)
										and: [(SystemNavigation default allCallsOn: #generateWith:using:) size <= 1].
isSqueak39 := Parser includesSelector: #pragmaSequence.

dot := FileDirectory default.
dotDot := dot directoryNamed: (dot class dirPathFor: dot fullName).
csdir := dotDot directoryNamed: 'changesets'.
getLatest := [:csn| | files splits ext |
	ext := '.cs'.
	files := csdir fileNamesMatching: csn, '*', ext.
	files isEmpty ifTrue: [ext := '.st'. files := csdir fileNamesMatching: csn, '*', ext].
	splits := files
			collect: [:file | csdir splitNameVersionExtensionFor: file]
			thenSelect: [:split | (split at: 1) = csn].
	splits := splits asSortedCollection: [:a :b | (a at: 2) < (b at: 2)].
	csn, ((splits last at: 2) = 0 ifTrue: [''] ifFalse: ['.', (splits last at: 2) asString]), ext].

loadCS := [:csn| "File-in a changeset.  FIlter out all Closure Usage Fixes.  These
				are redeclarations of block-local temps from method-level to block-local scope.
				We want these fixes but they're too various and hence too risky.  Better to
				hand-roll them specific to a given release."
			(csn beginsWith: 'Closure Usage Fixes')
				ifTrue:
					[Transcript cr; nextPutAll: 'Filtered out: '; nextPutAll: csn; cr; flush.]
				ifFalse:
					[Transcript cr; nextPutAll: csn; cr; flush.
					(csdir fileNamed: (getLatest value: csn)) fileIn]].
#(	'compiler-preload'
	'NeededFixes'
	'Debugger PC ParseNode Fixes'
	'PreRecompilationPatches'
	'DebuggerMethodMap'
	'MergeCompilers') do: loadCS.
hasTweak ifTrue:
	[loadCS value: 'MergeCompilers-TweakSide'.
	(Smalltalk includesKey: #CAssignmentTile) ifTrue:
		[loadCS value: 'MergeCompilers-CroquetSide'].
	(Smalltalk includesKey: #TScriptContext) ifTrue:
		[loadCS value: 'MergeCompilers-QwaqSide'].
	loadCS value: 'Recompilation-TweakSide'].
isSqueak39 ifTrue:
	[loadCS value: 'Recompilation-Squeak39Side'].
#(	'Recompilation'
	'BytecodeBrowsing'
	'Married Context Management Access'
	'Exeunt Old BlockClosure'
	'ParseNodeVisitorPattern'
	'Closure Compiler'
	'Closure Usage Fixes'
	'Closure Compiler Tests'
	'BlockContext Usage Fixes') do: loadCS.
isSqueak39 ifTrue:
	[loadCS value: 'BlockContext Usage Fixes-Squeak39Side'.
	 loadCS value: 'Closure Compiler-Squeak39Side'].
hasTweak
	ifTrue:
		[loadCS value: 'Closure Compiler-TweakSide'.
		(Smalltalk includesKey: #ScriptProcess) ifTrue:
			[loadCS value: 'Closure Usage Fixes-CroquetSide']]
	ifFalse:
		[loadCS value: 'Recompilation-SqueakSide'].
(Smalltalk includesKey: #BufferedAsyncFileReadStream) ifTrue:
	[loadCS value: 'Closure Usage Fixes-BARFS'].
canAndShouldDeleteGenerateWithUsing ifTrue:
	[(SystemNavigation default allCallsOn: #generateWith:using:) isEmpty
		ifTrue: [MethodNode removeSelector: #generateWith:using:]
		ifFalse: [nil inform: 'Did not remove MethodNode>>#generateWith:using:.  It has senders']].
Undeclared removeUnreferencedKeys.
(Undeclared keys anySatisfy: [:k| (previouslyUndeclared includes: k) not]) ifTrue:
	[nil inform: 'Undeclared has acquired some entries during the bootstrap!'.
	 newlyUndeclared := Dictionary new.
	 Undeclared associations do:
		[:assoc|
		(previouslyUndeclared includes: assoc key) ifFalse:
			[newlyUndeclared add: assoc]].
	 newlyUndeclared inspect].

"Now recompile using closures"
loadCS value: 'Compile Using Closures'.

"Now start to clean-up."
WeakArray restartFinalizationProcess.
MethodChangeRecord allInstancesDo:[:x| x noteNewMethod: nil].
Undeclared removeUnreferencedKeys.
Delay startTimerInterruptWatcher.
WorldState allInstancesDo:[:ws| ws convertAlarms; convertStepList].
(Workspace canUnderstand: #initializeBindings) ifTrue:
	[Workspace allInstancesDo:
		[:ws| ws initializeBindings]].
ExternalDropHandler initialize.
ScrollBar initializeImagesCache.
Vocabulary initialize.
Smalltalk garbageCollect.
GradientFillStyle initPixelRampCache.
Smalltalk
	at: #SokobanMorph
	ifPresent: [:sm| sm initFields].
Smalltalk
	at: #DebuggerMethodMap
	ifPresent: [:dmm| dmm voidMapCache].
Smalltalk
	at: #KClipboard
	ifPresent: [:kcb| kcb clearDefault].
Smalltalk
	at: #ServiceRegistry
	ifPresent: [:sr| sr rebuild].
(ProcessBrowser respondsTo: #registerWellKnownProcesses) ifTrue:
	[ProcessBrowser registerWellKnownProcesses].
self inform: 'Now make sure any Monticello browsers get closed.' withCRs.

Smalltalk forgetDoIts.
Smalltalk garbageCollect.
unboundMethods := CompiledMethod allInstances select:
					[:m|
					m methodClass isNil
					or: [m ~~ (m methodClass compiledMethodAt: m selector ifAbsent: nil)]].
unboundMethods := unboundMethods copyWithout: thisContext method.
unboundMethods notEmpty ifTrue:
	[(ToolSet inspect: unboundMethods) setLabel: 'Unbound Methods'].
self inform: 'Now follow the script at the end of this file')

"To eliminate the last of the old methods and block contexts do the following:
If viewing this file in a FileList, close the file list.  Close any browsers, message names, etc.  Just use a workspace (e.g. copy this text to a workspace).
Save and quit.
Restart.
Reopen the file.
If you have a morphic project to hand, enter it and exit it.  Otherwise, create a new morphic project, enter it, exit it and destroy it.
If in Squeak3.9 (or something with ServiceRegistry) evaluate
	(ServiceRegistry rebuild)
Evaluate
	(| unboundMethods |
	 Smalltalk forgetDoIts.
	 Smalltalk garbageCollect.
	 unboundMethods := CompiledMethod allInstances select:
						[:m|
						m methodClass isNil
						or: [m ~~ (m methodClass compiledMethodAt: m selector ifAbsent: nil)]].
	 unboundMethods := unboundMethods copyWithout: thisContext method.
	 unboundMethods notEmpty ifTrue:
		[(ToolSet inspect: unboundMethods) setLabel: 'Unbound Methods'])
You shoud end up with only 1 unbound method (the doit).  If you have more, Save and quit and try the doit again.  If unbound methods persist use e.g. "chase pointers" in the explorer on a method to track down references.  If you get down to about three give up and try tracking down BlockContexts (see next doit).
Evaluate
	(BlockContext allInstances)
This should be empty.  Again Save and quit.  Then repeat these steps and use e.g. chase pointers to track down remaining references."